/* fsign.c - OS signer for FreeFlash
 * Copyright (C) 2004-2005 Olivier Armand - <ola dot e-ml at wanadoo dot fr>
 * 
 * 2005-02-07:
 *  * Fixed a "comparison between signed and unsigned" warning
 *  * Added STRCMPI macro for *NIX compatibility
 *    (Godzil - godzil at godzil dot net)
 *  * Changed the default outfile name to <infile>_ff<extension> on Windows
 *  * Changed some error messages.
 * 
 * 2004-07-01:
 *  * Created
 */

#define WINDOWS /* define to allow using Win32 API */

/* *NIX compatibility for case-unsensitive comparision */
#ifdef WINDOWS
#define STRCMPI strcmpi
#else
#define STRCMPI strcasecmp
#endif

#ifdef WINDOWS
# include <windows.h>
#endif

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

const char TIFL[8] = "**TIFL**";
const char ErrFormat[] = "Input file is not a valid TIFL/TIB file.";
const char OSName[9] = { 8, 'b', 'a', 's', 'e', 'c', 'o', 'd', 'e' };
const unsigned char NewSignhead[65] = { 0x42, 0x80 };	/* and 63 zeros */
const unsigned char NewSigntail[] = { 0xFF, 0xFF };

#ifdef WINDOWS
#define TI68K_OS_PROMPT  "TI68k OS (*.tib;*.89u;*.9xu;*.v2u)\0*.tib;*.89u;*.9xu;*.v2u\0"
char strInFile[MAX_PATH] = "";
char strOutFile[MAX_PATH];
OPENFILENAME OpenFileName = { sizeof (OPENFILENAME), 0, 0,
  TI68K_OS_PROMPT,
  NULL, 0, 0, strInFile, MAX_PATH, NULL, 0, NULL,
  "Select the original OS...", OFN_FILEMUSTEXIST, 0, 0, NULL, 0, NULL, NULL
};
OPENFILENAME SaveFileName = { sizeof (OPENFILENAME), 0, 0,
  TI68K_OS_PROMPT, NULL, 0,
  0, strOutFile, MAX_PATH, NULL, 0, NULL, "Save the signed OS as...",
  OFN_OVERWRITEPROMPT, 0, 0, NULL, 0, NULL, NULL
};
#endif /* WINDOWS */

unsigned char *indata = NULL;
FILE *infile, *outfile;


#ifdef WINDOWS
void
FatalWindows (LPCTSTR msg)
{
  MessageBox (NULL, msg, "Error...", MB_OK | MB_ICONERROR);
}
#endif /* WINDOWS */

void
Fatal (const char *msg)
{
#ifdef WINDOWS
  FatalWindows (msg);
#else
  puts (msg);
#endif
  exit (0);
}

void
FatalMsg (const char *msg)
{
#ifdef WINDOWS
  FatalWindows ((LPCTSTR) _strerror (msg));
#else
  perror (msg);
#endif
  exit (0);
}

void
Quit (void)
{
  if (!indata)
    free (indata);
}

void
xfwrite (const void *src, size_t n)
{
  if (fwrite (src, sizeof (char), n, outfile) != n)
    Fatal ("Error while writing to the output file.");
}


#ifndef WINDOWS
int
main (int argc, char *argv[])
#else /* WINDOWS */
int WINAPI
WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine,
	 int nCmdShow)
#endif /* WINDOWS */
{
  unsigned char charbuf[4];
  int32_t os_size;
  int32_t main_size;
  int tib; /* != 0 if patching a .tib file */
  char *charptr;
  char *inname;
  char *outname;
  long insize;
  unsigned char *inptr;

  infile = outfile = NULL;
  atexit (Quit);

  /* Get the infile name */
#ifndef WINDOWS
  if (argc != 2 && argc != 3) {
    printf ("Usage: %s infile.xxu|.tib [outfile.xxu|.tib]\n", argv[0]);
    exit (0);
  }
  inname = argv[1];
  outname = (argc == 2 ? inname : argv[2]);
#else /* WINDOWS */
  if (!GetOpenFileName (&OpenFileName))
    exit (0);
  inname = strInFile;
  strcpy (strOutFile, strInFile);
  /* Add '_ff' at the end of the infile name for the default outfile name */
  charptr = strrchr(strOutFile, '.'); /* find the file extension */
  if (charptr != NULL) {
    memmove (charptr + sizeof("_ff") - sizeof((char)'\0'), charptr, strlen (charptr) + sizeof((char)'\0'));
    memcpy (charptr, "_ff", sizeof("_ff") - sizeof((char)'\0'));
  }
  if (!GetSaveFileName (&SaveFileName))
    exit (0);
  outname = strOutFile;
#endif /* WINDOWS */

  /* Read the infile (which may be the outfile) */
  if (!(infile = fopen (inname, "rb")))
    FatalMsg (inname);
  fseek (infile, 0, SEEK_END);
  inptr = indata = malloc (insize = ftell (infile));
  rewind (infile);
  if (!indata || fread (indata, sizeof (char), insize, infile) != (size_t)insize)
    FatalMsg ("Error while reading the input file.");
  fclose (infile);

  tib = ((charptr = strrchr (inname, '.')) && !STRCMPI (charptr, ".tib"));
  /* Check the format of the input file. */
  if (!tib) {
    if (memcmp (inptr, TIFL, sizeof (TIFL)))
      Fatal (ErrFormat);
    inptr += sizeof (TIFL) + 4 + 4; /* jump ? and the BCD coded date */
    if (memcmp (inptr, OSName, sizeof (OSName)))
      Fatal (ErrFormat);
    inptr += sizeof (OSName) + 23; /* jump a useless part of the header */
    /* calc type : 0x88 for 92+/V200, 0x98 for 89/Titanium  - var type : AMS */
    if ((*inptr != 0x88 && *inptr != 0x98))
      Fatal (ErrFormat);
    if (*++inptr != 0x23)
      Fatal (ErrFormat);
    inptr += 1 + 24; /* jump a useless part of the header */
    os_size = inptr[0] + (inptr[1] << 8) + (inptr[2] << 16) + (inptr[3] << 24); /* OS size to big endian */
    inptr += 4;
  }
  
  if (inptr[0] != 0x80 || inptr[1] != 0x0F) /* check the field ID of the certificate */
    Fatal (ErrFormat);
  inptr += 2;
  /* read the size of the field */
  main_size =
    (inptr[0] << 24) + (inptr[1] << 16) + (inptr[2] << 8) + inptr[3];
  inptr += 4 + main_size; /* jump the field */
  if (inptr[0] != 0x02 || inptr[1] != 0x0D) /* check the field ID (signature) and its size */
    Fatal (ErrFormat);
  if (inptr[2] != 64)
    Fatal ("The OS signature has already been modified.");
  inptr += 2; /* point to the size of the signature field */
  /* Check if os_size has a correct value.
   * 78: size of the header  /  1 + 64: size of the signature field and its size */
  if (!tib && (inptr - indata - 78 + 1 + 64 != os_size))
    Fatal (ErrFormat);

  /* Write the output file */
  if (!(outfile = fopen (outname, "wb")))
    FatalMsg (outname);
  xfwrite (indata, inptr - indata); /* Copy the OS */
  /* Write the new signature and its size */
  xfwrite (NewSignhead, sizeof (NewSignhead));
  xfwrite (NewSigntail, sizeof (NewSigntail));
  if (!tib) {
    /* New size for the OS */
    fseek (outfile, 74, SEEK_SET);
    os_size += 2;
    /* to little endian */
    charbuf[0] = (unsigned char) os_size;
    charbuf[1] = (unsigned char) (os_size >>= 8);
    charbuf[2] = (unsigned char) (os_size >>= 8);
    charbuf[3] = (unsigned char) (os_size >> 8);
    xfwrite (charbuf, 4);
  }
#ifdef WINDOWS
  MessageBox (NULL, "OS FSigned successfully.", "FSign: Success!",
	      MB_OK | MB_ICONINFORMATION);
#else
  puts ("OS FSigned successfully.");
#endif
  exit (0);
}
